// Copyright Derek C. Lazenby, 2019

unit TicTacToeMain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, LCLType;

type
  States = ( Playing, Won, Drawn ) ;
  LineRange = 1..8 ;
  LineSet = set of LineRange ;
  CellRange = 1..9 ;
  CellSet = set of CellRange ;

  { TTTTForm }

  TTTTForm = class(TForm)
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    Panel5: TPanel;
    Panel6: TPanel;
    Panel7: TPanel;
    Panel8: TPanel;
    Panel9: TPanel;
    procedure FormCreate(Sender: TObject);
    procedure PanelClick(Sender: TObject);
  private
    PlayerX : boolean;
    Count : integer;
    Status : States;
    XChoices,
    OChoices : CellSet ;
    const Winner : array [boolean] of char = ( 'O', 'X' ) ;
    Lines : array [LineRange] of CellSet = ( [1,2,3], [4,5,6], [7,8,9],
                                             [1,4,7], [2,5,8], [3,6,9],
                                             [1,5,9], [3,5,7] );
    Possible : array [CellRange] of LineSet = ( [1,4,7], [1,5], [1,6,8],
                                                [2,4], [2,5,7,8], [2,6],
                                                [3,4,8], [3,5], [3,6,7] );
    procedure ClearGrid;
  public

  end;

var
  TTTForm: TTTTForm;

implementation

{$R *.lfm}

{ TTTTForm }

procedure TTTTForm.FormCreate(Sender: TObject);
begin
  ClearGrid ; // well just init the variables actually!
end;

procedure TTTTForm.ClearGrid;
begin
  Panel1.Caption := '' ;
  Panel2.Caption := '' ;
  Panel3.Caption := '' ;
  Panel4.Caption := '' ;
  Panel5.Caption := '' ;
  Panel6.Caption := '' ;
  Panel7.Caption := '' ;
  Panel8.Caption := '' ;
  Panel9.Caption := '' ;
  XChoices := [] ;
  OChoices := [] ;
  Status := Playing ;
  PlayerX := true ;
  Count := 0 ;
end;

procedure TTTTForm.PanelClick(Sender: TObject);

  procedure CheckForWin ( var Choices : CellSet ) ;

    var Line   : LineRange ;
        Choice : CellRange ;

  begin
    Choice := TPanel (Sender).Tag ; // each panel (cell) has tag set to 1 to 9 respectively
    Choices := Choices + [Choice] ;
    if Count > 4 then
      for Line in Possible [Choice] do // only check relevant lines
        if Lines [Line] <= Choices then // check whole line !
        begin
          Status := Won ;
          break ;
        end ;
  end ;

  procedure GameOver ( Result : string ) ;
  begin
    if QuestionDlg ( 'Game Over', Result + #13#10#13#10'Play again?', mtConfirmation,
                    [mbyes, 'Yeah', mbno, 'Nope'], 0) = mrYes then
      ClearGrid
    else
      Close ;
  end ;

begin
  with TPanel ( Sender ) do
  begin
    if Caption = '' then
    begin
      Inc ( Count ) ;
      if PlayerX then
      begin
        Caption := 'X' ;
        CheckForWin ( XChoices ) ;
      end
      else
      begin
        Caption := 'O' ;
        CheckForWin ( OChoices ) ;
      end ;
      if ( Count = 9 ) and ( Status <> Won ) then
        Status := Drawn ;
      case Status of
      Playing: PlayerX := not PlayerX ;
      Won :    GameOver ( Winner [PlayerX] + ' Wins !!!' ) ;
      Drawn :  GameOver ( 'Game drawn.' ) ;
      end ;
    end;
  end ;
end;

end.

